// player.cpp: Beispielspieler fr Asteroids
// Harald Bgeholz / c't

//set if a>=


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <winsock2.h>
#include <process.h>

#include "player.h"

class CriticalSection
{
	volatile CRITICAL_SECTION cs;
	friend class CriticalSectionLocker;
public:
	CriticalSection()
	{
		InitializeCriticalSection((CRITICAL_SECTION*)&cs);
	}
	~CriticalSection()
	{
		DeleteCriticalSection((CRITICAL_SECTION*)&cs);
	}
private:
	CriticalSection(CriticalSection&);
	void operator=(CriticalSection& m);
};

class CriticalSectionLocker
{
	volatile CRITICAL_SECTION* cs;
public:
	CriticalSectionLocker(CriticalSection& m)
	{
		cs = &m.cs;
		EnterCriticalSection((CRITICAL_SECTION*)cs);
	}
	~CriticalSectionLocker()
	{
		LeaveCriticalSection((CRITICAL_SECTION*)cs);
	}
private:
	CriticalSectionLocker(CriticalSectionLocker&);
	void operator=(CriticalSectionLocker& m);
};


class ObjectsState
{
public:
	int StartTime, ValidUntilTime;
	int CollisionCauser;
	int DestroyedObject;
	int CountExtraSlots;
	Object Objects[NumObjects];
public:
	ObjectsState(int StartTime, int DestroyedObject, const GameRam& ram)
	{
		this->StartTime = StartTime;
		this->DestroyedObject = DestroyedObject;
		ValidUntilTime = MaxInt;
		CollisionCauser = -1;
		CountExtraSlots = ram.GetCountExtraSlots();
		memcpy(Objects, ram.GetObjects(), sizeof(Objects));
		for(int i = 0; i < NumObjects; i++)
			assert(i != DestroyedObject || Objects[i].Status <= 0);
	}
};

class ShotInfoList
{
	int MaxCollisionTime;
	std::vector<ShotInfo> ShotInfos;
public:
	ShotInfoList()
	{
		MaxCollisionTime = 0;

		ShotInfo Shot;
		Shot.Time = MaxInt;
		Shot.Dir = Dir_None;
		Shot.TargetDirection = 0;
		ShotInfos.push_back(Shot);
		ShotInfos.push_back(Shot);
	}

	const ShotInfo& operator [](int Index) const
	{
		return ShotInfos[Index];
	}

	bool operator <(const ShotInfoList& rhs) const
	{
/*		if(!Bad && rhs.Bad)
			return true;
		if(Bad && !rhs.Bad)
			return false;*/
		return MaxCollisionTime < rhs.MaxCollisionTime;
	}
	bool SameMovement(const ShotInfoList& rhs) const
	{
		//Gibt true zurck, wenn der Spielstatus um ShotInfos[0].Time+1 bei beiden gleich ist
		if(this == &rhs)
			return true;
		if(ShotInfos[0].Time == MaxInt && rhs.ShotInfos[0].Time == MaxInt)
			return true;
		return ShotInfos[0].Time == rhs.ShotInfos[0].Time &&
			   ShotInfos[0].TargetDirection == rhs.ShotInfos[0].TargetDirection &&
			   (ShotInfos[0].TargetDirection == ShotInfos[1].TargetDirection ? Dir_None : ShotInfos[1].Dir) == (rhs.ShotInfos[0].TargetDirection == rhs.ShotInfos[1].TargetDirection ? Dir_None : rhs.ShotInfos[1].Dir);
	}

	size_t size() const
	{
		return ShotInfos.size();
	}

	int GetMaxCollisionTime() const
	{
		return MaxCollisionTime;
	}

	void DeleteFirst()
	{
		ShotInfos.erase(ShotInfos.begin());
	}

	void AddShot(const ShotInfo& Shot, int Index, bool CrossedNextDir)
	{
		ShotInfos.insert(ShotInfos.begin()+Index, Shot);

		assert(Shot.CollisionTime >= MaxCollisionTime);
		MaxCollisionTime = Shot.CollisionTime;

		if(ShotInfos[Index+1].Time != MaxInt && CrossedNextDir)
		{
#ifdef PRINT_DEBUG_INFO
			printf("Aendere Richtung!\n");
#endif
			if(Shot.Dir == Dir_Left)
				ShotInfos[Index+1].Dir = Dir_Right;
			else
			{
				assert(Shot.Dir == Dir_Right);
				ShotInfos[Index+1].Dir = Dir_Left;
			}
		}

#ifdef PRINT_DEBUG_INFO
		printf("%sSchiesse mit Slot %i um %i auf %x! Kol: t=%i Dir: %i->%i\n",
			ShotInfos[Index+1].Time != MaxInt ? "[Sek]" : "",
			Shot.ShotIndex-PlayerShotMinIndex,
			Shot.Time,
			Shot.Target,
			Shot.CollisionTime,
			Shot.StartDirection,
			Shot.TargetDirection);
#endif
	}
};

std::vector<ObjectsState>* pObjectsStates = 0;

class ObjectsStateList
{
	std::vector<ObjectsState> ObjectsStates;
	GameRam Ram;
	const ShotInfoList& ShotInfos;
	int CurrentShotInfo;
public:
	ObjectsStateList(const GameRam& Ram, const ShotInfoList& ShotInfos)
		: ShotInfos(ShotInfos)
	{
		this->Ram = Ram;
		CurrentShotInfo = 0;
		ObjectsStates.push_back(ObjectsState(Ram.GetTime(), -1, Ram));
	}

	void Update(int NeededTime)
	{
		assert(pObjectsStates == 0);
		pObjectsStates = &ObjectsStates;
		try
		{
			while(Ram.GetTime() < NeededTime)
			{
				Ram.Move(ShotInfos[CurrentShotInfo], ShotInfos[CurrentShotInfo+1]);
				if(ShotInfos[CurrentShotInfo].Time == Ram.GetTime()-1)
					CurrentShotInfo++;
			}
		}
		catch(KIGameInvalidException)
		{
			pObjectsStates = 0;
			throw;
		}
		pObjectsStates = 0;
	}

	const ObjectsState& operator [](int Index) const
	{
		return ObjectsStates[Index];
	}

	size_t GetCountStates()
	{
		return ObjectsStates.size();
	}
};



#ifdef INCLUDE_WINDOW
HWND hwnd;
HBITMAP hBitmap;
HDC hdcMem;
int cxClient, cyClient;
const int xFont=9, yFont=16;
const int xChars = 0x100*2;
const int yChars = 50;

std::vector<GameRam> GameRamsHistory;
int CurrentGameRam = -1;
#endif

bool ValidateKIGame = false;


int ExplosionLifetime(char Status)
{
	assert(Status < 0);
	int Time = 0;
	while(Status < 0)
	{
		Status += (-Status >> 4) + 1;
		Time++;
	}
	return Time;
}

void GameRam::CalculateNextShot(const ShotInfoList& ShotInfos, int MinShootTime,
								std::multiset<ShotInfoList>& ShotInfosVariety) const 
{
	GameRam(*this).CalculateNextShot_(ShotInfos, MinShootTime, ShotInfosVariety);
}

void GameRam::CalculateNextShot_(const ShotInfoList& ShotInfos, int MinShootTime,
								std::multiset<ShotInfoList>& ShotInfosVariety)
{
	assert(ShotInfos[0].Time >= CurrentTime);


	//Wenn ein Asteroid explodiert, die neuen Asteroiden erscheinen oder das UFO
	//erscheint oder seine Richtung ndert, werden alle Objekte in ObjectsStates gespeichert
	//ObjectsStates simuliert parallel im eigenen GameRam so weit in die Zukunft, wie wir
	//die Daten in der KI brauchen.
	//So mssen wir nur einmal berechnen, wie sich die Asteroiden verhalten und knnen viele
	//Schsse simulieren, ohne stndig Move aufrufen zu mssen. V.a. die Kollisionsberechnung
	//wrde viel Zeit verschlingen, was wir so verhindern.
	ObjectsStateList ObjectsStates(*this, ShotInfos);


	int MinCollisionAtTime = MaxInt;//Der frheste Zeitpunkt, zu dem ein Asteroid abgeschossen werden kann (0=jetzt)

	int CurrentObjectsStates = 0;
	int CurrentShotInfo = 0;
	int FreeRotationTime = 0;//Wie oft wir uns gerade gedreht haben knnten (wird jeden Frame inkrementiert, wenn der Spieler nicht tot ist, Hyperspace ausfhrt...)
	bool RotationWasLocked = false;//Ob wir seit dem letzten Schuss gestorben sind
	int cShotsShot = 0;//Wie oft wir im aktuellen ObjectsState geschossen haben. Bentigen wir, da kein ObjectsStates erstellt wird, wenn geschossen wird
	for(int t = 0; t < MinCollisionAtTime; t++)
	{
		//Zu welchem Zeitpunkt schieen wir am besten? (t=0: jetzt)
		if(t)
		{
			if(ShotInfos[CurrentShotInfo].Time == CurrentTime)
			{
				ShipDirection = MaxShipDirection = ShotInfos[CurrentShotInfo].TargetDirection;
				FreeRotationTime = 0;
				RotationWasLocked = false;
			}

			if(NumPlayers &&
				Objects[PlayerIndex].Status >= 0 &&
				InvisibleTimer == 0)
				FreeRotationTime++;
			else
				RotationWasLocked = true;

			ObjectsStates.Update(CurrentTime+1);
			while(ObjectsStates[CurrentObjectsStates].ValidUntilTime == CurrentTime)
			{
				if(ValidateKIGame)
					assert(ObjectsStates[CurrentObjectsStates+1].CountExtraSlots >= ObjectsStates[CurrentObjectsStates].CountExtraSlots - cShotsShot, KIGameInvalidException());
				CurrentObjectsStates++;
				cShotsShot = 0;
			}
			assert(CurrentTime < ObjectsStates[CurrentObjectsStates].ValidUntilTime);

			byte OldShipDirection = ShipDirection;
			Move(ShotInfos[CurrentShotInfo], ShotInfos[CurrentShotInfo+1]);
			ShipDirection = MaxShipDirection = OldShipDirection;

			if(ShotInfos[CurrentShotInfo].Time == CurrentTime-1)
				CurrentShotInfo++;
		}

		if(ShotInfos[CurrentShotInfo].Time == CurrentTime)
			cShotsShot++;
		if(ValidateKIGame)
			assert(ObjectsStates[CurrentObjectsStates].CountExtraSlots - cShotsShot >= 0, KIGameInvalidException());

		if(CurrentTime < MinShootTime)
			continue;

		//Knnen wir schieen?
		if(FireHyperspaceToggle & 0x80 || InvisibleTimer)
			continue;

		//Schieen wir jetzt sowieso oder im nchsten Frame? (falls im nchsten drfen wir jetzt nicht schieen, weil wir sonst nicht im nchsten Frame schieen knnen)
		if(ShotInfos[CurrentShotInfo].Time == CurrentTime || ShotInfos[CurrentShotInfo].Time == CurrentTime+1)
			continue;

		if(ShotInfos[CurrentShotInfo].Time != MaxInt && ShotInfos[CurrentShotInfo].RotationWasLocked)
			continue;

		//Ist noch ein Slot frei?
		if(ObjectsStates[CurrentObjectsStates].CountExtraSlots - cShotsShot <= 0)
			continue;

		//Ist noch ein Slot frei?
		int NewShot = PlayerShotMaxIndex;
		while(Objects[NewShot].Status)
			if(NewShot-- == PlayerShotMinIndex)
				break;
		if(NewShot == PlayerShotMinIndex-1)
		{
			if(ValidateKIGame)
				assert(0, KIGameInvalidException());//Sollte wegen oberem if eigentlich nicht passieren
			continue;//Kein Slot frei
		}

		//Je nach Abschusszeitpunkt berlebt der Schuss unterschiedlich lange
		//int CollisionAtTime = Min(ShotLifetime-4 + (3-(Timer&3)) + t, MinCollisionAtTime);

		//Gibt an, wie wir uns zu diesem Zeitpunkt gedreht haben werden knnen
		int MinDeltaDir = -FreeRotationTime * RotationSpeed;
		int MaxDeltaDir = FreeRotationTime * RotationSpeed;
		if(ShotInfos[CurrentShotInfo].Time != MaxInt)
		{
			//Dafr sorgen, dass der nchste Schuss nach diesem Schuss erreicht werden kann
			const int PossibleRotation = (ShotInfos[CurrentShotInfo].Time-CurrentTime) * RotationSpeed;
			const int NeededRot = NeededRotation(ShipDirection, ShotInfos[CurrentShotInfo].TargetDirection, ShotInfos[CurrentShotInfo].Dir);
			MinDeltaDir = Max(MinDeltaDir, NeededRot - PossibleRotation);
			MaxDeltaDir = Min(MaxDeltaDir, NeededRot + PossibleRotation);
		}
		MinDeltaDir = Max(MinDeltaDir, -129/RotationSpeed*RotationSpeed); // In Tests hat sich herausgestellt, dass grere DeltaDirs selten benutzt werden und nichts bringen. Wir sparen uns also die Rechenzeit
		MaxDeltaDir = Min(MaxDeltaDir, 129/RotationSpeed*RotationSpeed);
		assert(MinDeltaDir%RotationSpeed == 0);
		assert(MaxDeltaDir%RotationSpeed == 0);
		//Wir prfen erst die betragsmig kleinen DeltaDirs, damit diese bei gleichem Kollisionszeitpunkt bevorzugt werden
		//Das gibt erstaunlich viele Punkte:)
		int MaxAbsDeltaDir = Max(MaxDeltaDir, -MinDeltaDir) / RotationSpeed * 2;
		for(int AbsDeltaDir = 0; AbsDeltaDir <= MaxAbsDeltaDir; AbsDeltaDir++)
		{
			int DeltaDir;
			if(AbsDeltaDir == 0)
				DeltaDir = 0;
			else if(AbsDeltaDir%2 == 0)
				DeltaDir = AbsDeltaDir / 2 * RotationSpeed;
			else
				DeltaDir = (AbsDeltaDir+1) / 2 * -RotationSpeed;
			if(DeltaDir < MinDeltaDir || DeltaDir > MaxDeltaDir)
				continue;
			//if(DeltaDir != MinDeltaDir && DeltaDir != MaxDeltaDir)
			//	continue;///////////////////

			SimulateShot(ObjectsStates, ShotInfos, CurrentObjectsStates, CurrentShotInfo, cShotsShot,
				         t, NewShot, DeltaDir, MinCollisionAtTime, ShotInfosVariety, RotationWasLocked);

			//Hier simulieren wir einen Schuss in t Frames bei einer Schiffsdrehung von DeltaDir (relativ zur Richtung beim letzten Schuss)

/*			word StartX = Objects[PlayerIndex].PositionX;
			word StartY = Objects[PlayerIndex].PositionY;
			char VelocityX = Objects[PlayerIndex].VelocityX;
			char VelocityY = Objects[PlayerIndex].VelocityY;
			ShootLocationFromAngle(ShipDirection+DeltaDir,
								   StartX,
								   StartY,
								   VelocityX,
								   VelocityY);

			int CurrentObjectsStates2 = CurrentObjectsStates;
			int CurrentShotInfo2 = CurrentShotInfo;
			int cShotsShot2 = cShotsShot;
			for(int t2 = t; t2 < CollisionAtTime; t2++)
			{
				int CurrentTime2 = CurrentTime+t2-t;
				ObjectsStates.Update(CurrentTime2+1);
				//Wir simulieren die Bewegung des Schusses

				if(ShotInfos[CurrentShotInfo2].Time == CurrentTime2)
				{
					CurrentShotInfo2++;
					cShotsShot2++;
					assert(ObjectsStates[CurrentObjectsStates2].CountExtraSlots - cShotsShot2 >= 0);
					if(ObjectsStates[CurrentObjectsStates2].CountExtraSlots - cShotsShot2 <= 0)
						break;
				}

				while(CurrentTime2 == ObjectsStates[CurrentObjectsStates2].ValidUntilTime &&
					NewShot < ObjectsStates[CurrentObjectsStates2].CollisionCauser)
				{
					CurrentObjectsStates2++;
					cShotsShot2 = 0;
				}
				assert(CurrentTime2 <= ObjectsStates[CurrentObjectsStates2].ValidUntilTime);

				const Object* Objects2 = ObjectsStates[CurrentObjectsStates2].Objects;

				for(int Object = SaucerIndex-2; Object >= 0; Object--)
				{
					//Trifft der Schuss ein Objekt?
					if(Objects2[Object].Status <= 0)
						continue;
					bool collision = AreObjectsNear(NewShot,
													MoveLinearX(StartX, VelocityX, t2-t+1),
													MoveLinearY(StartY, VelocityY, t2-t+1),
													ShotLifetime/4,//wird ignoriert
						                            Object,
													MoveLinearX(Objects2[Object].PositionX, Objects2[Object].VelocityX, CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1),
													MoveLinearY(Objects2[Object].PositionY, Objects2[Object].VelocityY, CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1),
													Objects2[Object].Status);
					if(!collision)
						continue;
					//Wir treffen das Objekt!!

					//Wir schieen aber nicht, falls es gleich (noch im gleichen Frame) von einem anderen Schuss
					//getroffen wird.
					//Dies kann passieren, wenn unser Schuss einen hheren Index bekommt als der, der auf dieses
					//Objekt unterwegs ist, und wir das Objekt deshalb trotzdem erreichen knnen, obwohl wir
					//offensichtlich spter schieen. (Schsse mit hherem Index werden von Asteroids zuerst abgearbeitet)
					//Dieser Fall tritt aber zugegebenermaen sehr selten ein..
					int CurrentObjectsStates3 = CurrentObjectsStates2;
					while(CurrentTime2 == ObjectsStates[CurrentObjectsStates3++].ValidUntilTime)
						if(ObjectsStates[CurrentObjectsStates3].DestroyedObject == Object)
						{
#ifdef PRINT_DEBUG_INFO
							printf("Vermeide Schuss mit Slot %i auf %x um %i\n", NewShot-PlayerShotMinIndex, Object, CurrentTime);
#endif
							break;
						}
					if(CurrentTime2 == ObjectsStates[CurrentObjectsStates3-1].ValidUntilTime)
					{
						t2 = MaxInt;//Dieses DeltaDir kommt nicht in Frage
						break;
					}

//					while(CurrentObjectsStates3 < ObjectsStates.GetCountStates())
//						assert(ObjectsStates[CurrentObjectsStates3++].DestroyedObject != Object);

					//Wenn wir einen groen Asteroiden abknallen mssen wir darauf achten, dass es genug
					//Slots fr die zwei kleinen gibt.
					if(!(Objects2[Object].Status & 0x01))
					{
						int cFreeAsteroidSlots = 0;
						for(int i = AsteroidMinIndex; i <= AsteroidMaxIndex; i++)
							if(Objects2[i].Status == 0 ||
								Objects2[i].Status < 0 && AsteroidLifetime(Objects2[i].Status) <= CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1)
							{
								if(++cFreeAsteroidSlots == 2)
									break;
							}
						if(cFreeAsteroidSlots != 2)
						{
							//Wenn wir dieses Asteroiden abknallen wrden, wrden keine 2 kleinere Asteroiden entstehen
							//Wir vermeiden also diese Richtung zu diesem Zeitpunkt
#ifdef PRINT_DEBUG_INFO
							printf("Vermeide Schuss mit Slot %i auf %x um %i, da zu wenig Slots da sind\n", NewShot-PlayerShotMinIndex, Object, CurrentTime);
#endif
							t2 = MaxInt;
							break;
						}
					}

					if(CurrentTime2 < ShotInfos.GetMaxCollisionTime())
					{
#ifdef PRINT_DEBUG_INFO
						printf("Zu frueher Schuss\n");
#endif
						t2 = MaxInt;
						break;
					}


					bool CrossedNextDir = false;
					if(ShotInfos[CurrentShotInfo].Time != MaxInt && DeltaDir)
					{
						for(int i = 0; i != DeltaDir; i += (DeltaDir>0?RotationSpeed:-RotationSpeed))
							if(byte(i+ShipDirection) == ShotInfos[CurrentShotInfo].TargetDirection)
							{
								CrossedNextDir = true;
								break;
							}
					}

					CollisionAtTime = t2;
					MinCollisionAtTime = t2;

					ShotInfo BestShotInfo;
					BestShotInfo.Time = CurrentTime;
					if(DeltaDir > 0)
						BestShotInfo.Dir = Dir_Left;
					else if(DeltaDir < 0)
						BestShotInfo.Dir = Dir_Right;
					else
						BestShotInfo.Dir = Dir_None;
					BestShotInfo.TargetDirection = ShipDirection+DeltaDir;
					BestShotInfo.CollisionTime = CurrentTime2;
				
					BestShotInfo.ShotIndex = NewShot;
					BestShotInfo.PositionX = MoveLinearX(StartX, VelocityX, t2-t+1);
					BestShotInfo.PositionY = MoveLinearY(StartY, VelocityY, t2-t+1);
					BestShotInfo.Target = Object;
					BestShotInfo.StartX = StartX;
					BestShotInfo.StartY = StartY;
					BestShotInfo.VelocityX = VelocityX;
					BestShotInfo.VelocityY = VelocityY;
					BestShotInfo.TargetPositionX = MoveLinearX(Objects2[Object].PositionX, Objects2[Object].VelocityX, CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1);
					BestShotInfo.TargetPositionY = MoveLinearY(Objects2[Object].PositionY, Objects2[Object].VelocityY, CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1);
					BestShotInfo.TargetStatus = Objects2[Object].Status;

					ShotInfoList ShotInfos2 = ShotInfos;
					ShotInfos2.AddShot(BestShotInfo, CurrentShotInfo, CrossedNextDir);
					OriginalRam.CheckForErrors(ShotInfos2, BestShotInfo.CollisionTime+20);
					ShotInfosVariety.insert(ShotInfos2);
					break;
				}
				if(t2 == MaxInt)
					break;
				while(CurrentTime2 == ObjectsStates[CurrentObjectsStates2].ValidUntilTime)
				{
					CurrentObjectsStates2++;
					cShotsShot2 = 0;
				}
				assert(CurrentTime2 < ObjectsStates[CurrentObjectsStates2].ValidUntilTime);
			}*/
		}
	}
}

void GameRam::SimulateShot(ObjectsStateList& ObjectsStates, const ShotInfoList& ShotInfos,
						   const int CurrentObjectsStates, const int CurrentShotInfo, const int cShotsShot,
						   const int t,
						   const int ShotIndex, const int DeltaDir, int& MinCollisionAtTime,
						   std::multiset<ShotInfoList>& ShotInfosVariety, bool RotationWasLocked)
{
	word StartX = Objects[PlayerIndex].PositionX;
	word StartY = Objects[PlayerIndex].PositionY;
	char VelocityX = Objects[PlayerIndex].VelocityX;
	char VelocityY = Objects[PlayerIndex].VelocityY;
	ShootLocationFromAngle(ShipDirection+DeltaDir,
						   StartX,
						   StartY,
						   VelocityX,
						   VelocityY);

	//Je nach Abschusszeitpunkt berlebt der Schuss unterschiedlich lange
	int CollisionAtTime = Min(ShotLifetime-4 + (3-(Timer&3)) + t, MinCollisionAtTime);

	int CurrentObjectsStates2 = CurrentObjectsStates;
	int CurrentShotInfo2 = CurrentShotInfo;
	int cShotsShot2 = cShotsShot;
	for(int t2 = t; t2 < CollisionAtTime; t2++)
	{
		int CurrentTime2 = CurrentTime+t2-t;
		ObjectsStates.Update(CurrentTime2+1);
		//Wir simulieren die Bewegung des Schusses

		if(ShotInfos[CurrentShotInfo2].Time == CurrentTime2)
		{
			CurrentShotInfo2++;
			cShotsShot2++;
			if(ValidateKIGame)
				assert(ObjectsStates[CurrentObjectsStates2].CountExtraSlots - cShotsShot2 >= 0, KIGameInvalidException());
			if(ObjectsStates[CurrentObjectsStates2].CountExtraSlots - cShotsShot2 <= 0)
				break;
		}

		while(CurrentTime2 == ObjectsStates[CurrentObjectsStates2].ValidUntilTime &&
			ShotIndex < ObjectsStates[CurrentObjectsStates2].CollisionCauser)
		{
			CurrentObjectsStates2++;
			cShotsShot2 = 0;
		}
		assert(CurrentTime2 <= ObjectsStates[CurrentObjectsStates2].ValidUntilTime);

		const Object* Objects2 = ObjectsStates[CurrentObjectsStates2].Objects;

		for(int Object = SaucerIndex; Object >= AsteroidMinIndex; Object--)
		{
			if(Object == PlayerIndex)//Wir knnen zwar vom eigenen Schuss getroffen werden, allerdings nur mit Hyperspace. Und den benutzen wir eh nicht.
				continue;
			//Trifft der Schuss ein Objekt?
			if(Objects2[Object].Status <= 0)
				continue;
			bool collision = AreObjectsNear(ShotIndex,
											MoveLinearX(StartX, VelocityX, t2-t+1),
											MoveLinearY(StartY, VelocityY, t2-t+1),
											ShotLifetime/4,//wird ignoriert
				                            Object,
											MoveLinearX(Objects2[Object].PositionX, Objects2[Object].VelocityX, CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1),
											MoveLinearY(Objects2[Object].PositionY, Objects2[Object].VelocityY, CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1),
											Objects2[Object].Status);
			if(!collision)
				continue;
			//Wir treffen das Objekt!!

			//Wir schieen aber nicht, falls es gleich (noch im gleichen Frame) von einem anderen Schuss
			//getroffen wird.
			//Dies kann passieren, wenn unser Schuss einen hheren Index bekommt als der, der auf dieses
			//Objekt unterwegs ist, und wir das Objekt deshalb trotzdem erreichen knnen, obwohl wir
			//offensichtlich spter schieen. (Schsse mit hherem Index werden von Asteroids zuerst abgearbeitet)
			//Dieser Fall tritt aber zugegebenermaen sehr selten ein..
			int CurrentObjectsStates3 = CurrentObjectsStates2;
			while(CurrentTime2 == ObjectsStates[CurrentObjectsStates3++].ValidUntilTime)
				if(ObjectsStates[CurrentObjectsStates3].DestroyedObject == Object)
				{
#ifdef PRINT_DEBUG_INFO
					printf("Vermeide Schuss mit Slot %i auf %x um %i\n", ShotIndex-PlayerShotMinIndex, Object, CurrentTime);
#endif
					break;
				}
			if(CurrentTime2 == ObjectsStates[CurrentObjectsStates3-1].ValidUntilTime)
			{
				t2 = MaxInt;//Dieses DeltaDir kommt nicht in Frage
				break;
			}

			//Wenn wir einen groen Asteroiden abknallen mssen wir darauf achten, dass es genug
			//Slots fr die zwei kleinen gibt.
			if(!(Objects2[Object].Status & 0x01))
			{
				int cFreeAsteroidSlots = 0;
				for(int i = AsteroidMinIndex; i <= AsteroidMaxIndex; i++)
					if(Objects2[i].Status == 0 ||
						Objects2[i].Status < 0 && ExplosionLifetime(Objects2[i].Status) <= CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1)
					{
						if(++cFreeAsteroidSlots == 2)
							break;
					}
				if(cFreeAsteroidSlots != 2)
				{
					//Wenn wir dieses Asteroiden abknallen wrden, wrden keine 2 kleinere Asteroiden entstehen
					//Wir vermeiden also diese Richtung zu diesem Zeitpunkt
#ifdef PRINT_DEBUG_INFO
					printf("Vermeide Schuss mit Slot %i auf %x um %i, da zu wenig Slots da sind\n", ShotIndex-PlayerShotMinIndex, Object, CurrentTime);
#endif
					t2 = MaxInt;
					break;
				}
			}

			if(CurrentTime2 < ShotInfos.GetMaxCollisionTime())
			{
//#ifdef PRINT_DEBUG_INFO
//				printf("Zu frueher Schuss\n");
//#endif
				t2 = MaxInt;
				break;
			}

			MinCollisionAtTime = CollisionAtTime = t2;

			ShotInfo BestShotInfo;
			BestShotInfo.Time = CurrentTime;
			if(DeltaDir > 0)
				BestShotInfo.Dir = Dir_Left;
			else if(DeltaDir < 0)
				BestShotInfo.Dir = Dir_Right;
			else
				BestShotInfo.Dir = Dir_None;
			BestShotInfo.TargetDirection = ShipDirection+DeltaDir;
			BestShotInfo.CollisionTime = CurrentTime2;
			BestShotInfo.RotationWasLocked = RotationWasLocked;

			BestShotInfo.ShotIndex = ShotIndex;
			BestShotInfo.PositionX = MoveLinearX(StartX, VelocityX, t2-t+1);
			BestShotInfo.PositionY = MoveLinearY(StartY, VelocityY, t2-t+1);
			BestShotInfo.Target = Object;
			BestShotInfo.StartX = StartX;
			BestShotInfo.StartY = StartY;
			BestShotInfo.VelocityX = VelocityX;
			BestShotInfo.VelocityY = VelocityY;
			BestShotInfo.TargetPositionX = MoveLinearX(Objects2[Object].PositionX, Objects2[Object].VelocityX, CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1);
			BestShotInfo.TargetPositionY = MoveLinearY(Objects2[Object].PositionY, Objects2[Object].VelocityY, CurrentTime2-ObjectsStates[CurrentObjectsStates2].StartTime+1);
			BestShotInfo.TargetStatus = Objects2[Object].Status;
			BestShotInfo.StartDirection = ShipDirection;

			bool CrossedNextDir = false;
			if(ShotInfos[CurrentShotInfo].Time != MaxInt && DeltaDir)
			{
				for(int i = 0; i != DeltaDir; i += (DeltaDir>0?RotationSpeed:-RotationSpeed))
					if(byte(ShipDirection+i) == ShotInfos[CurrentShotInfo].TargetDirection)
					{
						CrossedNextDir = true;
						break;
					}
			}


			ShotInfoList ShotInfos2 = ShotInfos;
			ShotInfos2.AddShot(BestShotInfo, CurrentShotInfo, CrossedNextDir);
			//OriginalRam.CheckForErrors(ShotInfos2, BestShotInfo.CollisionTime+20);
			ShotInfosVariety.insert(ShotInfos2);
			break;
		}
		if(t2 == MaxInt)
			break;
		while(CurrentTime2 == ObjectsStates[CurrentObjectsStates2].ValidUntilTime)
		{
			CurrentObjectsStates2++;
			cShotsShot2 = 0;
		}
		assert(CurrentTime2 < ObjectsStates[CurrentObjectsStates2].ValidUntilTime);
	}
}

void GameRam::CheckForErrors(const ShotInfoList& ShotInfos, int UntilTime)
{
	//Zuknftige Schritte simulieren, um schon jetzt Fehler abzufangen, die spter auftreten knnen (erleichtert Debugging)
	GameRam OriginalRam = *this;
	//bool ErrorsOccured = false;
	//pErrorsOccured = &ErrorsOccured;
	int CurrentShotInfo = 0;
	while(CurrentTime != UntilTime)
	{
		Move(ShotInfos[CurrentShotInfo], ShotInfos[CurrentShotInfo+1]);
		if(ShotInfos[CurrentShotInfo].Time == CurrentTime-1)
			CurrentShotInfo++;
	}
	*this = OriginalRam;
	/*pErrorsOccured = 0;
	if(ErrorsOccured)
	{
		printf("Fehler...:(\n");
		while(true)
			;
		MessageBeep(0);
		//exit(0);
		// CalcDirInfos(pShotInfo, pDeltaDir, CalcSecondShoot, FixedShootAtTime, FixedShootAtDir, FixedKey);
	}*/
}


//Prft, ob das Objekt CollisionCauser mit irgendwas kollidiert und fhrt
//entsprechende Aktionen aus
void GameRam::CheckCollisions(byte CollisionCauser,
							  int FixedTarget,
							  std::stack<MovePosition<GameRam>>* pMovePositions)//sub_69F0
{
	if(Objects[CollisionCauser].Status <= 0 && FixedTarget == -1)
		return;
	int y = SaucerIndex;
	if(CollisionCauser < PlayerShotMinIndex)//CollisionCauser ist kein Schuss vom Spieler
		y--;// kann also nicht mit dem UFO kollidieren
	if(CollisionCauser == PlayerIndex)
		y--;//Spieler kann nicht mit sich selbst kollidieren
	if(FixedTarget != -1)
		y = FixedTarget;
	do
	{
		if(Objects[y].Status <= 0)
		{
			assert(y != FixedTarget);
			continue;
		}

		if(FixedTarget != y)
		{
			int cCollisionCombinations = 0, cNoCollisionCombinations = 0;
			Object ColNewCollisionCauser(Objects[CollisionCauser].Status);
			Object ColNewTarget(Objects[y].Status);
			Object NoColNewCollisionCauser(Objects[CollisionCauser].Status);
			Object NoColNewTarget(Objects[y].Status);

			for(int ax = 0; ax < Objects[CollisionCauser].cX; ax++)
			for(int ay = 0; ay < Objects[CollisionCauser].cY; ay++)
			for(int bx = 0; bx < Objects[y].cX; bx++)
			for(int by = 0; by < Objects[y].cY; by++)
	//		for(int ax = Objects[CollisionCauser].cX-1; ax >= 0; ax--)
	//		for(int ay = Objects[CollisionCauser].cY-1; ay >= 0; ay--)
	//		for(int bx = Objects[y].cX-1; bx >= 0; bx--)
	//		for(int by = Objects[y].cY-1; by >= 0; by--)
			{
				bool Near = AreObjectsNear(CollisionCauser,
										   Objects[CollisionCauser].PX[ax],
										   Objects[CollisionCauser].PY[ay],
										   Objects[CollisionCauser].Status,
										   y,
										   Objects[y].PX[bx],
										   Objects[y].PY[by],
										   Objects[y].Status);
				if(Near)
					cCollisionCombinations++;
				else
					cNoCollisionCombinations++;
				if(Near)
				{
					ColNewCollisionCauser.AddLocation(Objects[CollisionCauser].PX[ax], Objects[CollisionCauser].PY[ay],
													  Objects[CollisionCauser].VX[ax], Objects[CollisionCauser].VY[ay]);
					ColNewTarget.AddLocation(Objects[y].PX[bx], Objects[y].PY[by],
											 Objects[y].VX[bx], Objects[y].VY[by]);
				}
				else
				{
					NoColNewCollisionCauser.AddLocation(Objects[CollisionCauser].PX[ax], Objects[CollisionCauser].PY[ay],
														Objects[CollisionCauser].VX[ax], Objects[CollisionCauser].VY[ay]);
					NoColNewTarget.AddLocation(Objects[y].PX[bx], Objects[y].PY[by],
											   Objects[y].VX[bx], Objects[y].VY[by]);
				}
			}
			
			if(cNoCollisionCombinations && cCollisionCombinations == 0)
			{
				//Es sind eine oder mehrere Kombinationen, die alle nicht kollidieren
				//->Kein Entscheidungsproblem
				continue;
			}
			else if(cNoCollisionCombinations == 0 && cCollisionCombinations)
			{
				//Es gibt nur Kollisionen
				//->Kein Entscheidungsproblem
			}
			else if(cNoCollisionCombinations && cCollisionCombinations)
			{
				assert(ColNewCollisionCauser.IsValid());
				assert(ColNewTarget.IsValid());
				assert(NoColNewCollisionCauser.IsValid());
				assert(NoColNewTarget.IsValid());

				MovePosition<GameRam> NewMovePosition(MPT_ObjectCollision, true);
				NewMovePosition.CollisionCauser = CollisionCauser;
				NewMovePosition.Target = y;
				NewMovePosition.Game = *this;

				NewMovePosition.Game.Objects[CollisionCauser] = ColNewCollisionCauser;
				NewMovePosition.Game.Objects[y] = ColNewTarget;
				pMovePositions->push(NewMovePosition);

				Objects[CollisionCauser] = NoColNewCollisionCauser;
				Objects[y] = NoColNewTarget;

				printf("Entscheide fuer keine Kollisionen um %i  No: %i %i\n", CurrentTime, cNoCollisionCombinations, cCollisionCombinations);
				continue;//Keine Kollision
			}
			else
				assert(0);
		}

		//Wir kollidieren!

		if(OnCollision(CollisionCauser, y, pMovePositions) && pObjectsStates)
		{
			//An den Objekten hat sich etwas wichtiges gendert. Fr die KI speichern.
			pObjectsStates->back().ValidUntilTime = CurrentTime;
			pObjectsStates->back().CollisionCauser = CollisionCauser;
			pObjectsStates->push_back(ObjectsState(CurrentTime+1, y, *this));
		}
		return;//Nur eine Kollision pro Objekt
	}while(y--);
}

void GameRam::ResetAsteroids(std::stack<MovePosition<GameRam>>* pMovePositions)//sub_7168
{
	int x = AsteroidMaxIndex;
	if(!NewLevelStartCountdown)
	{
		if(Objects[SaucerIndex].Status)
			return;

		if(RandomSeedDistribution.GetNumSetBits() != 1)
		{
			//Wir werden gleich neue Asteroiden erstellen, kennen den Seed aber noch nicht
			ChooseRandomSeed(pMovePositions, MPT_ResetAsteroids);
		}

		Objects[SaucerIndex].SetVelocity(0, 0);

		if(Level < 10)
			Level++;

		NumAsteroidsStart += 2;
		if(NumAsteroidsStart >= 0x0B)
			NumAsteroidsStart = 0x0B;
		NumAsteroids = NumAsteroidsStart;
		for(int i = 0; i < NumAsteroids; i++)
		{
			Objects[x].Status = (Rand()&0x18) | 0x04;

			//sub_7203
			{
				byte a = Rand() & 0x8F;
				if(char(a) < 0)
					a |= 0xF0;
				for(int i = 0; i < Objects[x].cX; i++)
				{
					Objects[x].VX[i] = a;
					ClampVelocity(Objects[x].VX[i]);
				}

				Rand(); Rand(); Rand();

				a = Rand() & 0x8F;
				if(char(a) < 0)
					a |= 0xF0;
				for(int i = 0; i < Objects[x].cY; i++)
				{
					Objects[x].VY[i] = a;
					ClampVelocity(Objects[x].VY[i]);
				}
			}

			byte a = Rand();
			bool c;
			ShiftRight(a, c);
			a &= 0x1F;
			if(c)
			{
				if(a >= 0x18)
					a &= 0x17;
				for(int i = 0; i < Objects[x].cY; i++)
					HighByte(Objects[x].PY[i]) = a;
				for(int i = 0; i < Objects[x].cX; i++)
					Objects[x].PX[i] = 0;
			}
			else
			{
				for(int i = 0; i < Objects[x].cX; i++)
					HighByte(Objects[x].PX[i]) = a;
				for(int i = 0; i < Objects[x].cY; i++)
					Objects[x].PY[i] = 0;
			}
			Objects[x].RemoveEqualities();
			x--;
		}
		SaucerCountdown = 0x7F;
		//Var(RAM_2FC) = 0x30;
	}
	
	while(x >= AsteroidMinIndex)
		Objects[x--].Status = 0;
	
	if(pObjectsStates)
	{
		pObjectsStates->back().ValidUntilTime = CurrentTime;
		pObjectsStates->back().CollisionCauser = -1;
		pObjectsStates->push_back(ObjectsState(CurrentTime+1, -1, *this));
	}
}

void GameRam::UpdateSaucer(std::stack<MovePosition<GameRam>>* pMovePositions)//sub_6B93
{
	assert(TimerBitsKnown >= 2);

	if(Timer & 3 || Objects[SaucerIndex].Status < 0)
		return;

	if(Objects[SaucerIndex].Status == 0)
	{
		if(NumPlayers && Objects[PlayerIndex].Status <= 0)
			return;

		if(RandomSeedDistribution.GetNumSetBits() != 1 &&
			SaucerCountdown == 1 &&
			(AsteroidDestroyedTimer <= 1 || (NumAsteroids != 0 && NumAsteroids < Level)))
		{
			//Wir werden gleich ein UFO erstellen, kennen aber den Seed noch nicht.
			ChooseRandomSeed(pMovePositions, MPT_UpdateSaucer);
		}

		if(AsteroidDestroyedTimer)
			AsteroidDestroyedTimer--;
		if(--SaucerCountdown)
			return;
		SaucerCountdown = 0x12;
		if(AsteroidDestroyedTimer && (NumAsteroids == 0 || NumAsteroids >= Level))
			return;

		//Saucer erstellen
		if(SaucerCountdownStart >= 38)
			SaucerCountdownStart -= 6;
		byte RandomSeed2;
		byte a = Rand(&RandomSeed2);

		//Setze y Position (y Geschwindigkeit wird nicht angerhrt)
		/*			bool c;
		ShiftRight(a, c);
		RotateRight(LowByte(Objects[SaucerIndex].PositionY), c);
		ShiftRight(a, c);
		RotateRight(LowByte(Objects[SaucerIndex].PositionY), c);
		ShiftRight(a, c);
		RotateRight(LowByte(Objects[SaucerIndex].PositionY), c);*/

		for(int i = 0; i < Objects[SaucerIndex].cY; i++)
		{
			Objects[SaucerIndex].PY[i] = ((Objects[SaucerIndex].PY[i]>>3) & 0x1F) | ((a&7) << 5);
			HighByte(Objects[SaucerIndex].PY[i]) = a;
			HighByte(Objects[SaucerIndex].PY[i]) >>= 3;
			if(HighByte(Objects[SaucerIndex].PY[i]) >= 0x18)
				HighByte(Objects[SaucerIndex].PY[i]) &= 0x17;
		}

		//x Komponente
		Objects[SaucerIndex].PositionX = RandomSeed2 & 0x40 ? 0 : 0x1FFF;
		Objects[SaucerIndex].VelocityX = RandomSeed2 & 0x40 ? 16 : -16;
		Objects[SaucerIndex].cX = 1;

		//Status
		Objects[SaucerIndex].Status = 2;
		if(char(SaucerCountdownStart) >= 0 && (Score%100000 >= 30000 || (SaucerCountdownStart>>1) <  Rand()))
			Objects[SaucerIndex].Status = 1;

		Objects[SaucerIndex].RemoveEqualities();

		if(pObjectsStates)
		{
			//Der KI bescheid sagen, dass sich an den Objekten etwas gendert hat
			pObjectsStates->back().ValidUntilTime = CurrentTime;
			pObjectsStates->back().CollisionCauser = MaxInt;//Wird vor den Kollisionen ausgefhrt
			pObjectsStates->push_back(ObjectsState(CurrentTime+1, -1, *this));
		}
		return;
	}
	//UFO lebt:

	if(SaucerCountdown == 1 &&
		(!NumPlayers || !InvisibleTimer))
	{
		//Wir werden gleich schieen
		if(RandomSeedDistribution.GetNumSetBits() != 1)
		{
			//Wir kennen den Seed noch nicht.
			ChooseRandomSeed(pMovePositions, MPT_UpdateSaucer);
		}
		if(!Objects[SaucerIndex].IsFixed())
		{
			//Die Berechnung des Schusses setzt voraus, dass die Position genau bekannt ist.
			//Man knnte die Berechnung auch so gestalten, dass sie mit der Ungenauigkeit klar
			//kommt, das wre aber sehr aufwendig und der Geschwindigkeitsvorteil wre eher klein.
			assert(Objects[SaucerIndex].cX == 1);

			MovePosition<GameRam> NewMovePosition(MPT_UpdateSaucer, false);
			NewMovePosition.Game = *this;
			NewMovePosition.Game.Objects[SaucerIndex].RemoveY(Objects[SaucerIndex].cY-1);
			pMovePositions->push(NewMovePosition);

			Objects[SaucerIndex].Decide(0, Objects[SaucerIndex].cY-1);
		}
	}

	//Richtung hin und wieder ndern
	if(TimerBitsKnown < 7)
	{
		if((Timer & (0xFF >> (8-TimerBitsKnown))) == 0)//Sind alle bekannten Bits 0?
		{
			//Wenn unser Timer falsch geht knnte das UFO jetzt die Richtung ndern
			MovePosition<GameRam> NewMovePosition(MPT_UpdateSaucer, false);
			NewMovePosition.Game = *this;
			NewMovePosition.Game.Timer = 0;
			NewMovePosition.Game.TimerBitsKnown = 7;
			pMovePositions->push(NewMovePosition);
			printf("Saucer knnte jetzt Richtung ndern. um %i\n", CurrentTime);
		}
	}
	else if(byte(Timer << 1) == 0)
	{
		if(RandomSeedDistribution.GetNumSetBits() != 1)
		{
			//Wir werden gleich die Richtung ndern, kennen aber den Seed noch nicht.
			ChooseRandomSeed(pMovePositions, MPT_UpdateSaucer);
		}

		char OldVelocityY = Objects[SaucerIndex].VelocityY;
		const char DirectionTable[] = {-16, 0, 0, 16};
		Objects[SaucerIndex].SetVelocityY(DirectionTable[Rand()&3]);

		if(pObjectsStates && OldVelocityY != Objects[SaucerIndex].VelocityY)
		{
			//Der KI bescheid sagen, dass sich an den Objekten etwas gendert hat
			pObjectsStates->back().ValidUntilTime = CurrentTime;
			pObjectsStates->back().CollisionCauser = MaxInt;//Wird vor den Kollisionen ausgefhrt
			pObjectsStates->push_back(ObjectsState(CurrentTime+1, -1, *this));
		}
	}

	if(NumPlayers && InvisibleTimer)
		return;

	if(--SaucerCountdown)
		return;

	//Wir feuern mal einen Schuss
	SaucerCountdown = 0x0A;

	byte ShotDirection;
	if(Objects[SaucerIndex].Status != 2)
	{
		byte a = Objects[SaucerIndex].VelocityX;
		bool c = a >= 0x80;
		RotateRight(a, c);
		byte tmpC = a;
		a = LowByte(Objects[PlayerIndex].PositionX);
		c = true;
		Subtract(a, LowByte(Objects[SaucerIndex].PositionX), c);
		byte tmpB = a;
		a = HighByte(Objects[PlayerIndex].PositionX);
		Subtract(a, HighByte(Objects[SaucerIndex].PositionX), c);
		ShiftLeft(tmpB, c);
		RotateLeft(a, c);
		ShiftLeft(tmpB, c);
		RotateLeft(a, c);
		c = true;
		Subtract(a, tmpC, c);
		byte x = a;

		a = Objects[SaucerIndex].VelocityY;
		c = a >= 0x80;
		RotateRight(a, c);
		tmpC = a;
		a = LowByte(Objects[PlayerIndex].PositionY);
		c = true;
		Subtract(a, LowByte(Objects[SaucerIndex].PositionY), c);
		tmpB = a;
		a = HighByte(Objects[PlayerIndex].PositionY);
		Subtract(a, HighByte(Objects[SaucerIndex].PositionY), c);
		ShiftLeft(tmpB, c);
		RotateLeft(a, c);
		ShiftLeft(tmpB, c);
		RotateLeft(a, c);
		c = true;
		Subtract(a, tmpC, c);
		byte y = a;

		ShotDirection = loc_76F0(x, y);

		a = Rand();
		c = Score%100000 >= 35000;
		a &= c ? 0x87 : 0x8F;
		if(char(a) < 0)
			a |= c ? 0x78 : 0x70;
		Add(ShotDirection, a, c);
	}
	else
		ShotDirection = Rand();

	int NewShot = SaucerShotMaxIndex;
	while(Objects[NewShot].Status)
		if(NewShot-- == SaucerShotMinIndex)
			return;

	assert(Objects[SaucerIndex].IsFixed());
	Objects[NewShot] = Objects[SaucerIndex];
	Objects[NewShot].Status = ShotLifetime/4;

	ShootLocationFromAngle(ShotDirection, Objects[NewShot].PositionX, Objects[NewShot].PositionY, Objects[NewShot].VelocityX, Objects[NewShot].VelocityY);
}


std::vector<GameRam> VRamGames;

#pragma pack(1)

struct KeysPacket
{
	char Signature[6];
	byte Keys;
	byte Ping;     // wird vom Server bei nchster Gelegenheit zurckgeschickt. Fr Latenzmessung.
};
#pragma pack()

#ifdef USE_NETWORK_THREAD

struct NetBufferItem
{
	FramePacket Frame;
	bool FrameRecieved;
	byte Keys;
	byte Ping;
	//int FramePing2;
};

#ifdef _DEBUG
const int NetBufferSize = 40*60;
#else
const int NetBufferSize = 20*60;
#endif
const int NetBufferKIFutureSteps = 3*60;
volatile NetBufferItem NetBuffer[NetBufferSize];

volatile int NetBufferPosition = 0;
volatile int KINetBufferPosition = 0;
volatile int NetDelay = 0;
volatile bool NetworkThreadExit = false;
CriticalSection NetBufferPositionLock;

#endif

sockaddr_in ServerAddress;
SOCKET ServerSocket;

#ifdef USE_NETWORK_THREAD

void NetworkThread(void* p)
{
	SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);

	KeysPacket Keys;
	memcpy(Keys.Signature, "ctmame", 6);

	byte CurrentPing = 0;

	//std::ofstream log("NetThreadLog.txt");

#ifdef WRITE_FRAMES
	std::ofstream out("AsteroidsSave.dat", std::ios::binary);
#elif defined READ_FRAMES
	std::ifstream in("AsteroidsSave.dat", std::ios::binary);
#endif

	bool First = true;

	while(!NetworkThreadExit)
	{
		int NetBufferPositionCopy = NetBufferPosition;
		volatile NetBufferItem& Item = NetBuffer[NetBufferPositionCopy];
		volatile NetBufferItem& LastItem = NetBuffer[(NetBufferPositionCopy-1+NetBufferSize) % NetBufferSize];

		Item.Ping = CurrentPing++;
		Keys.Keys = Item.Keys;

		/*if(GetForegroundWindow() == hwnd)
		{
			Keys.Keys = 0;
			if(GetAsyncKeyState('Q') < 0)
				Keys.Keys |= Key_Left;
			if(GetAsyncKeyState('W') < 0)
				Keys.Keys |= Key_Right;
			if(GetAsyncKeyState('I') < 0)
				Keys.Keys |= Key_Thrust;
			if(GetAsyncKeyState('O') < 0)
				Keys.Keys |= Key_Fire;
			if(GetAsyncKeyState(VK_SPACE) < 0)
				Keys.Keys |= Key_Hyperspace;
			Item.Keys = Keys.Keys;
		}*/

		Keys.Ping = Item.Ping;
	//	log<<"["<<int(NetBufferPositionCopy)<<"] Sende Keys: "<<int(Keys.Keys)<<"  Ping: "<<int(Keys.Ping)<<"    ";
		int r = sendto(ServerSocket, (char*)&Keys, sizeof(KeysPacket), 0,
					   (sockaddr*)&ServerAddress, sizeof(ServerAddress));
		if(r != sizeof(KeysPacket))
			printf("Fehler beim Senden (%i)\n", WSAGetLastError());

		if(First)
		{
			char Buffer[38] = {0};
			strcpy_s(Buffer, 38, "ctnameHelmut Buhler");
			r = sendto(ServerSocket, Buffer, sizeof(Buffer), 0,
				(sockaddr*)&ServerAddress, sizeof(ServerAddress));
			if(r != sizeof(Buffer))
				printf("Fehler beim Senden des Namens (%i)\n", WSAGetLastError());
		}

//		log.flush();


#ifdef INCLUDE_WINDOW
		//Belegung des Ringpuffers anzeigen
		//Das sollte man hier direkt vor recv machen, da recv eh blockiert und wir hier
		//wahrscheinlich etwas Zeit haben
		HDC hdc = GetDC(hwnd);
		RECT rc;
		SetRect(&rc,
			0,
			cyClient-10,
			(NetBufferPosition-KINetBufferPosition+NetBufferSize) % NetBufferSize * cxClient / NetBufferSize,
			cyClient);
		FillRect(hdc, &rc, (HBRUSH)GetStockObject(BLACK_BRUSH));
		ReleaseDC(hwnd, hdc);
#endif

RecievePacketAgain:
		r = recv(ServerSocket, (char*)&Item.Frame, sizeof(FramePacket), 0);
		if(r != sizeof(FramePacket))
		{
			r -= 2;
			if(r > 127)
				r = 127;
			if(r < 0)
				r = 0;
			char Buffer[128];
			memcpy(Buffer, (void*)&Item.Frame, r);
			Buffer[r] = 0;
			printf("Server: \"%s\" (Code: %i)\n", Buffer, WSAGetLastError());
			break;
		}

#ifdef WRITE_FRAMES
		out.write((char*)&Item.Frame, sizeof(Item.Frame));
		out.flush();
#elif defined READ_FRAMES
		in.read((char*)&Item.Frame, sizeof(Item.Frame));
		if(in.fail())
			return;
#endif

//		log<<"Empfange Ping: "<<int(Item.Frame.Ping)<<" No: "<<int(Item.Frame.FrameNo)<<"\n";

		if(!First && Item.Frame.FrameNo != byte(LastItem.Frame.FrameNo + 1))
		{
			char RelNo = char(byte(Item.Frame.FrameNo - byte(LastItem.Frame.FrameNo + 1)));
			MessageBeep(0);
			if(RelNo > 0)
			{
				printf("%i Frames vom Server verloren. Eventuell stuerzt das Programm bald ab, da Informationen fehlen..\n", RelNo);
				//Versuchen wir zu retten, was zu retten ist...

				Item.FrameRecieved = false;
				Item.Frame.FrameNo = LastItem.Frame.FrameNo+1;
				int NewPosition = (NetBufferPositionCopy+RelNo) % NetBufferSize;
				while(NetBufferPositionCopy != NewPosition)
				{
					NetBufferPositionCopy = (NetBufferPositionCopy+1) % NetBufferSize;
					NetBuffer[NetBufferPositionCopy].Frame.Ping = LastItem.Frame.Ping;
#ifdef INCLUDE_EXTRA_FRAME_INFO
					NetBuffer[NetBufferPositionCopy].Frame.LastRecvKeys = LastItem.Frame.LastRecvKeys;
#endif
					NetBuffer[NetBufferPositionCopy].Frame.FrameNo = ++Item.Frame.FrameNo;
					NetBuffer[NetBufferPositionCopy].FrameRecieved = false;
					NetBuffer[NetBufferPositionCopy].Ping = CurrentPing++;
					NetBuffer[NetBufferPositionCopy].Keys = Item.Keys;//Wir gehen mal davon aus, dass der Server kein neues Paket von uns empfangen hat. Falls doch und sich die Keys unterscheiden, wird das Spiel asynchron:(
				}
				((FramePacket&)NetBuffer[NetBufferPositionCopy].Frame) = ((FramePacket&)Item.Frame);///////////////////////
				Item.Frame.FrameNo = LastItem.Frame.FrameNo+1;
				Item.Frame.Ping = LastItem.Frame.Ping;
#ifdef INCLUDE_EXTRA_FRAME_INFO
				Item.Frame.LastRecvKeys = LastItem.Frame.LastRecvKeys;
#endif
			}
			else
			{
				int CorrectPosition = (NetBufferPositionCopy+RelNo+NetBufferSize) % NetBufferSize;
				if(NetBuffer[CorrectPosition].FrameRecieved)
					printf("Frame von vor %i Frames ist vom Server doppelt angekommen\n", -RelNo);
				else
					printf("Frame von vor %i Frames ist vom Server in falscher Reihenfolge angekommen\n", -RelNo);
				((FramePacket&)NetBuffer[CorrectPosition].Frame) = ((FramePacket&)Item.Frame);///////////////////////
				NetBuffer[CorrectPosition].FrameRecieved = true;
				goto RecievePacketAgain;
			}
		}
		NetBuffer[NetBufferPositionCopy].FrameRecieved = true;

		First = false;


		if((NetBufferPositionCopy+1) % NetBufferSize == KINetBufferPosition)
		{
			printf("KI rechnet zu lange: Netzwerkpuffer ist voll. Beende.\n");
			break;
		}

		if(NetDelay)
			Sleep(NetDelay);


		{
			CriticalSectionLocker l = NetBufferPositionLock;
			NetBufferPosition = (NetBufferPositionCopy+1) % NetBufferSize;
		}
	}
	NetworkThreadExit = true;
}
#endif

void MoveVRamGames(byte LastKeys, const word* CurrentVRam, const byte* LastRam,
				   int& LastVRamScore, int& LastVRamNumShips, GameRam* pKIGame)
{
	int NewVRamScore = -1, NewVRamNumShips;
	assert(!VRamGames.empty());
	std::vector<GameRam> NewVRamGames;
	for(size_t i = 0; i < VRamGames.size(); i++)
	{
//		printf("VRam Move %i\n", i);
		VRamGames[i].VRamMove(LastKeys, CurrentVRam, LastRam, NewVRamGames, LastVRamScore, LastVRamNumShips, NewVRamScore, NewVRamNumShips, pKIGame);
	}

	if(VRamGames[0].NumPlayers)
	{
		assert(!NewVRamGames.empty());
		VRamGames = NewVRamGames;

		bool OredSomething = false;
		for(size_t i = 0; i < VRamGames.size(); i++)
		for(size_t j = 0; j < i; j++)
			if(VRamGames[j].CanBeOredWith(VRamGames[i]))
			{
				VRamGames[j] |= VRamGames[i];
				OredSomething = true;
				VRamGames.erase(VRamGames.begin()+i);
				i--;
				break;
			}
		if(OredSomething)
			for(size_t i = 0; i < VRamGames.size(); i++)
				VRamGames[i].RandomSeedDistribution.UpdateInternalData();
	}

	LastVRamScore = NewVRamScore;
	LastVRamNumShips = NewVRamNumShips;
}


std::multiset<ShotInfoList> ShotInfosVariety;
void Player::Run()
{
//	ShellExecuteA(0, "open", "C:\\C++-Programme\\Asteroids\\vcmamed.exe", "", 0, SW_SHOW);
//	FramePacket last_frame;
	byte changed[0x100] = {0};
//	char prevframe = 0;
	int t = 0;
//	char current_ping=0;
//	char last_ping=127;
//	char last_last_ping=32;
//	__int64 Score = 0;
//	byte StartSeed = 0;
//	DecisionList Decisions;
//	std::map<int, GameRam> DecisionGameRams;
//	word LastVRam[512];
//	byte LastRam[0x300];

	VRamGames.resize(1);
	VRamGames[0].StartNewGame();

	ShotInfosVariety.insert(ShotInfoList());

#ifdef USE_NETWORK_THREAD
	//std::ofstream log("KILog.txt");
	int LatencyPosition = -1;
#else

#ifdef WRITE_FRAMES
	std::ofstream out("AsteroidsSave.dat", std::ios::binary);
#elif defined READ_FRAMES
	std::ifstream in("AsteroidsSave.dat", std::ios::binary);
#endif

	KeysPacket Keys;
	memcpy(Keys.Signature, "ctmame", 6);
	Keys.Keys = Keys.Ping = 0;
	FramePacket LastFrame;

#endif

	int LastVRamScore = -1, LastVRamNumShips;

	GameRam KIGame;
	KIGame.CurrentTime = 0;


	int FutureShots = 13;
	//int MaxNeeded = 0;

	memset(&ServerAddress, 0, sizeof(ServerAddress));
	ServerAddress.sin_family = AF_INET;
	ServerAddress.sin_port = htons(1979);
//	ServerAddress.sin_addr.s_addr = inet_addr("127.0.0.1");
	ServerAddress.sin_addr.s_addr = inet_addr("193.99.145.178");
//	ServerAddress.sin_addr.s_addr = inet_addr("192.168.178.4");

#ifdef USE_NETWORK_THREAD
	for(int i = 0; i < NetBufferSize; i++)
	{
		NetBuffer[i].Keys = 0x80;
	}

	_beginthread(NetworkThread, 0, 0);
#endif


	while(true)
	{
#ifdef INCLUDE_WINDOW
		MSG msg;
		while(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
		{
			if(msg.message==WM_QUIT)
				return;//Eigentlich mssten wir hier den Netzwerkthread noch sauber beenden, aber ich berlass das dann doch lieber Windows:)
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		if(CurrentGameRam != -1)
		{
			//Wenn wir uns einen alten Frame anschauen pausieren wir mal die KI,
			//damit das Fenster schneller reagiert
			WaitMessage();
			continue;
		}
#endif

#ifdef USE_NETWORK_THREAD
		if(NetworkThreadExit)
		{
			printf("Erreichter Score: %i\n", VRamGames[0].GetScore());
			break;
		}
		const int NetBufferPositionCopy = NetBufferPosition;
		while(KINetBufferPosition != NetBufferPositionCopy)
		{
			const int KINetBufferPositionCopy = KINetBufferPosition;
			volatile NetBufferItem& Current = NetBuffer[KINetBufferPositionCopy];
			if(t >= 1)
			{
				volatile NetBufferItem& Last = NetBuffer[(KINetBufferPositionCopy-1+NetBufferSize) % NetBufferSize];

				int Latency = byte(Current.Ping - Last.Frame.Ping);
				LatencyPosition = (KINetBufferPositionCopy - Latency + NetBufferSize) % NetBufferSize;
				byte LastKeys = NetBuffer[LatencyPosition].Keys;

				assert(NetBuffer[LatencyPosition].Ping == Last.Frame.Ping);
	#ifdef INCLUDE_EXTRA_FRAME_INFO
				assert(t < 10 || LastKeys == Last.Frame.LastRecvKeys);
	#endif
				assert(Current.Frame.FrameNo == byte(Last.Frame.FrameNo + 1));//Das wird vom Netzwerkthread garantiert

				if(t > 60 && Current.Frame.Ping != byte(Last.Frame.Ping + 1))
				{
					if(char(byte(Last.Frame.Ping + 1 - Current.Frame.Ping)) > 0)
						printf("Paket ist nicht zur richtigen Zeit beim Server angekommen. %i zu spaet\n", byte(Last.Frame.Ping + 1 - Current.Frame.Ping));
					else
						printf("Paket ist nicht zur richtigen Zeit beim Server angekommen. %i zu frueh\n", byte(Current.Frame.Ping - Last.Frame.Ping - 1));
				}

				//printf("Eigener Frame: Hash: %x\n", Game.GetHash());
				//printf("Move Key: %i\n", LastKeys);
				MoveVRamGames(LastKeys,
					Current.FrameRecieved ? (const word*)Current.Frame.VRam : 0,
#ifdef INCLUDE_EXTRA_FRAME_INFO
					Current.FrameRecieved ? (const byte*)Last.Frame.Ram : 0,
#else
					0,
#endif
					LastVRamScore, LastVRamNumShips, &KIGame);///////////////////cast

				//GameRamsHistory.push_back(Game);
			}
			KINetBufferPosition = (KINetBufferPositionCopy+1) % NetBufferSize;
			t++;
		}



		if(LatencyPosition != -1 && t >= 3 && KIGame.CurrentTime)
		{
			GameRam KIGame(KIGame);
			//Latency = current_ping - frame.ping + 2;
			//printf("Latenz: %i\n", Latency);
			int CurrentPosition = LatencyPosition;
			int BeginPosition = NetBufferPosition;
			while(CurrentPosition != (BeginPosition+20) % NetBufferSize)
			{
				KIGame.Move(NetBuffer[CurrentPosition].Keys);
				CurrentPosition = (CurrentPosition+1) % NetBufferSize;
			}
			//printf("Berechne neuen Schuss ab %i\n", KIGame.CurrentTime);

			ShotInfoList OldShotInfos = *ShotInfosVariety.begin();
			while(OldShotInfos[0].Time < KIGame.CurrentTime)
			{
				//printf("Loesche Schuss um %i\n", OldShotInfos[0].Time);
				OldShotInfos.DeleteFirst();
			}
			assert(!ValidateKIGame);
			ValidateKIGame = true;
TryAgain:
			ShotInfosVariety.clear();
			try
			{
				KIGame.CalculateNextShot(OldShotInfos, 0, ShotInfosVariety);
			}
			catch(KIGameInvalidException)
			{
				//MessageBeep(0);
				//assert(OldShotInfos[0].Time != MaxInt);
				assert(ValidateKIGame);
				OldShotInfos = ShotInfoList();
				ValidateKIGame = false;
				printf("Loesche Schuesse wegen Fehler\n", OldShotInfos[0].Time);
				goto TryAgain;
				throw;///////////////////
			}
			ValidateKIGame = false;

			ShotInfoList& NewShotInfos = *ShotInfosVariety.begin();
			int CurrentShotInfo = 0;

			{
			CriticalSectionLocker l = NetBufferPositionLock;

//			log<<"Neuer Schuss\n";
			for(int i = 0; i < NetBufferKIFutureSteps; i++)
			{
				byte Keys = KIGame.GetKeyFromShotInfo(NewShotInfos[CurrentShotInfo], NewShotInfos[CurrentShotInfo+1]);
				if((CurrentPosition-BeginPosition+NetBufferSize) % NetBufferSize <= (NetBufferPosition-BeginPosition+NetBufferSize) % NetBufferSize)
				//if(CurrentPosition <= NetBufferPosition)//Funktioniert leider nicht im Ring
				{
					if(NetBuffer[CurrentPosition].Keys == Keys)
					{
						printf("Schuss zu spaet, aber zum Glueck gleich.\n");
//						log<<"["<<CurrentPosition<<"] Ist schon: Key: "<<int(NetBuffer[CurrentPosition].Keys)<<"  Zeit: "<<KIGame.CurrentTime<<"\n";
					}
					else
					{
						printf("Schuss zu spaet und falsch!!\n");
						ShotInfosVariety.clear();
						ShotInfosVariety.insert(OldShotInfos);
						break;
					}
				}
				else
				{
					NetBuffer[CurrentPosition].Keys = Keys;
			//		log<<"["<<CurrentPosition<<"] Key: "<<int(NetBuffer[CurrentPosition].Keys)<<"  Zeit: "<<KIGame.CurrentTime<<"\n";
				}
				KIGame.Move(Keys);
				if(NewShotInfos[CurrentShotInfo].Time == KIGame.CurrentTime-1)
					CurrentShotInfo++;
				if(NewShotInfos[CurrentShotInfo].Time == MaxInt)
					break;
				CurrentPosition = (CurrentPosition+1) % NetBufferSize;
				assert(CurrentPosition != BeginPosition);
			}
			}
			//log.flush();
		}
#ifdef INCLUDE_WINDOW
		InvalidateRect(hwnd, 0, false);
#endif

#else


		//Paket senden
		int r = sendto(ServerSocket, (char*)&Keys, sizeof(KeysPacket), 0,
					   (sockaddr*)&ServerAddress, sizeof(ServerAddress));
		if(r != sizeof(KeysPacket))
			printf("Fehler beim Senden (%i)\n", WSAGetLastError());
	
/*		//Auf neues Paket warten
		fd_set readfds, exceptfds;
		FD_ZERO(&readfds);
		FD_ZERO(&exceptfds);
		FD_SET(ServerSocket, &readfds);
		FD_SET(ServerSocket, &exceptfds);
		select(0, &readfds, 0, &exceptfds, 0);*/

		FramePacket Frame;

#ifndef READ_FRAMES
		r = recv(ServerSocket, (char*)&Frame, sizeof(FramePacket), 0);
		if(r != sizeof(FramePacket))
		{
			r -= 2;
			if(r > 127)
				r = 127;
			if(r < 0)
				r = 0;
			char Buffer[128];
			memcpy(Buffer, (void*)&Frame, r);
			Buffer[r] = 0;
			printf("Server: \"%s\" (Code: %i)\n", Buffer, WSAGetLastError());
			return;
		}
#endif

#ifdef WRITE_FRAMES
		out.write((char*)&Frame, sizeof(Frame));
		out.flush();
#elif defined READ_FRAMES
		in.read((char*)&Frame, sizeof(Frame));
		if(in.fail())
			return;
#endif


		const int MaxSameShotInfos = 10;
/*		if(ShotInfosVariety.empty())
		{
			ShotInfosVariety.insert(ShotInfoList());
#ifndef USER_CONTROL
			for(int i = 0; i <= FutureShots; i++)
			{
				std::multiset<ShotInfoList> NewShotInfosVariety;
				int Index = 0;
				for(std::multiset<ShotInfoList>::iterator j = ShotInfosVariety.begin();
					j != ShotInfosVariety.end();
					j++)
				{
					if(++Index == MaxSameShotInfos)
						break;
					gameram.CalculateNextShot(*j, 0, NewShotInfosVariety);
				}
				assert(!NewShotInfosVariety.empty());
				ShotInfosVariety = NewShotInfosVariety;
			}
#endif
		}*/

		/*while(ShotInfos.size() < 4)
		{
			if(ShotInfos.size() == 3 && ShotInfos[0].CollisionTime == MaxCollisionTime)
			{
				int ShootTime = ShotInfos[0].Time;
				gameram.CalculateNextShot(ShotInfos, MaxCollisionTime, 0);
				int MinMaxCollisionTime = MaxCollisionTime;
				int BestDelay;

				//printf("Test\n");
				for(int i = 1; i <= 2 ; i++)
				{
					std::vector<ShotInfo> ShotInfos2 = EmptyShotInfos;
					int MaxCollisionTime2 = 0;
					gameram.CalculateNextShot(ShotInfos2, MaxCollisionTime2, ShootTime+i);
					if(MaxCollisionTime2 >= MinMaxCollisionTime)
						break;
					gameram.CalculateNextShot(ShotInfos2, MaxCollisionTime2, 0);
					if(MaxCollisionTime2 < MinMaxCollisionTime)
					{
						ShotInfos = ShotInfos2;
						MinMaxCollisionTime = MaxCollisionTime2;
						BestDelay = i;
					}
				}
				//printf("TestEnde\n");
#ifdef PRINT_DEBUG_INFO
				if(MinMaxCollisionTime < MaxCollisionTime)
					printf("Juhuu. Spare %i mit der Verzoegerung %i\n", MaxCollisionTime - MinMaxCollisionTime, BestDelay);
#endif
				MaxCollisionTime = MinMaxCollisionTime;
			}
			else
				gameram.CalculateNextShot(ShotInfos, MaxCollisionTime, 0);
		}*/

		if(t && Frame.FrameNo != byte(LastFrame.FrameNo+1))
			printf("%i Frames verloren.\n", byte(Frame.FrameNo - byte(LastFrame.FrameNo+1)));

#ifdef INCLUDE_EXTRA_FRAME_INFO
		//Frame.Ping = Frame.LastRecvKeys;/////////////////////////
		assert(Frame.LastRecvKeys == Frame.Ping);
#endif

		//printf("Move Key: %x Ping: %x\n", keys_packets[i].keys & 0x1f, keys_packets[i].ping);
		//printf("Frame Hash: %x\n", gameram.Hash());


		if(t > 1)
		{
			MoveVRamGames(LastFrame.Ping, Frame.VRam,
#ifdef INCLUDE_EXTRA_FRAME_INFO
				LastFrame.Ram,
#else
				0,
#endif
				LastVRamScore, LastVRamNumShips, &KIGame);
/*			int NewVRamScore = -1, NewVRamNumShips;
			std::vector<GameRam> NewVRamGames;
			for(size_t i = 0; i < VRamGames.size(); i++)
			{
				//printf("VRam Move %i\n", i);
				VRamGames[i].VRamMove(LastFrame.Ping, Frame.VRam, LastFrame.Ram, NewVRamGames, LastVRamScore, LastVRamNumShips, NewVRamScore, NewVRamNumShips, &KIGame);
			}

			if(VRamGames[0].NumPlayers)
			{
				assert(!NewVRamGames.empty());
				VRamGames = NewVRamGames;

				bool OredSomething = false;
				for(size_t i = 0; i < VRamGames.size(); i++)
				for(size_t j = 0; j < i; j++)
					if(VRamGames[j].CanBeOredWith(VRamGames[i]))
					{
						VRamGames[j] |= VRamGames[i];
						OredSomething = true;
						VRamGames.erase(VRamGames.begin()+i);
						i--;
						break;
					}
				if(OredSomething)
					for(size_t i = 0; i < VRamGames.size(); i++)
						VRamGames[i].RandomSeedDistribution.UpdateInternalData();
			}

			assert(NewVRamScore != -1);
			LastVRamScore = NewVRamScore;
			LastVRamNumShips = NewVRamNumShips;*/
		}


/*		int TargetTime = gameram.CurrentTime+1;
		while(gameram.CurrentTime != TargetTime)
		{
			try
			{
				assert(CurrentVRamTime - gameram.CurrentTime < DecisionTime);
				word* VRam = VRams[(CurrentVRam - (CurrentVRamTime - gameram.CurrentTime) + DecisionTime) % DecisionTime];
				DecisionAdded = false;
				gameram.Move_(KeysPacket(), ShotInfo(), frame.ram, VRam, Decisions);
			}
			catch(VRamInvalidException)
			{
				if(DecisionAdded)
				{
					printf("Entscheidung um %i\n", GameRam::last_ram.CurrentTime);
					DecisionGameRams[GameRam::last_ram.CurrentTime] = GameRam::last_ram;
				}
				printf("Aendere Entscheidung\n");
				assert(!Decisions.empty());
				while(Decisions.begin()->second.NextDecision())
				{
					Decisions.erase(Decisions.begin());
					assert(!Decisions.empty());
					printf("Springe zu Zeitpunkt %i zurueck\n", Decisions.begin()->first.GetTime());
				}
				assert(DecisionGameRams.find(Decisions.begin()->first.GetTime()) != DecisionGameRams.end());
				assert(CurrentVRamTime - Decisions.begin()->first.GetTime() < DecisionTime);
				gameram = DecisionGameRams[Decisions.begin()->first.GetTime()];
				if(Decisions.find(DecisionKey::RandomSeed(gameram.CurrentTime)) != Decisions.end() &&
					Decisions[DecisionKey::RandomSeed(gameram.CurrentTime)].GetTriesLeft() != gameram.RandomSeedTriesLeft)
				{
					assert(gameram.RandomSeedTriesLeft > 0);
					gameram.RandomSeedTriesLeft--;
					assert(Decisions[DecisionKey::RandomSeed(gameram.CurrentTime)].GetTriesLeft() == gameram.RandomSeedTriesLeft);
					gameram.Rand();
					DecisionGameRams[Decisions.begin()->first.GetTime()] = gameram;
					printf("Naechster Seed. Noch %i\n", gameram.RandomSeedTriesLeft);
				}
				continue;
				throw;//////////////////////
			}
			if(DecisionAdded)
			{
				printf("Entscheidung um %i\n", GameRam::last_ram.CurrentTime);
				DecisionGameRams[GameRam::last_ram.CurrentTime] = GameRam::last_ram;
			}
		}
		gameram.Check(frame.ram, keys_packets[i], ShotInfo(), false, server_ip, sd, Decisions);				*/
/*				}
			}
			else if(t != 1)
				printf("%i Paket(e) verloren\n", int(frame.frameno - prevframe+char(1)));
			if(prevframe-frame.frameno >= 0 && prevframe-frame.frameno < 10)
			{
				printf("Paket ignoriert\n");
				continue;
			}
			//gameram.Check(frame.ram, keys_packets[0], ShotInfo(), true, 0, 0);
			last_ping = frame.ping;
			//if(frame.frameno != prevframe+1)
			//	printf("Frame Hash: %x\n", gameram.Hash());
		}
		else*/
/*		{
		}*/



		/*BitBlt(hdcMem, 0, yFont*2, xChars*xFont, (yChars-2)*yFont, hdcMem, 0, yFont, SRCCOPY);
		for(int i=0; i<0x100; i++)
		{
			const int Start = 0x200;
			if(t > 20 && frame.ram.GetRawRam()[i+Start] != last_frame.ram.GetRawRam()[i+Start])
				changed[i] = 255;
			SetTextColor(hdcMem, RGB(changed[i],0,0));
			if(changed[i]) changed[i]--;
			char Buffer[3];
			sprintf_s(Buffer, 3, "%2.2x", frame.ram.GetRawRam()[i+Start]);
			TextOut(hdcMem, i*xFont*2, yFont, Buffer, 2);
		}*/
		InvalidateRect(hwnd, 0, false);

//		last_frame = frame;


		//if(updating) InterpretScreen(frame, game);

#if 1
		if(KIGame.CurrentTime)
		{
/*			Latency = current_ping - frame.ping + 2;
			//printf("Latenz: %i\n", Latency);
			for(int i = 0; i < Latency; i++)
				KIGame.Move(keys_packets[(t+i-Latency)%10]);*/
			ShotInfoList ShotInfos;

#ifndef READ_FRAMES
			KIGame.Move(LastFrame.Ping);
			KIGame.Move(Frame.Ping);

			ShotInfosVariety.clear();
			KIGame.CalculateNextShot(ShotInfos, 0, ShotInfosVariety);
			ShotInfos = *ShotInfosVariety.begin();
#endif
			/*ShotInfosVariety.clear();
			KIGame.CalculateNextShot(ShotInfos, 0, ShotInfosVariety);
			ShotInfos = *ShotInfosVariety.begin();*/

			Keys.Keys = KIGame.GetKeyFromShotInfo(ShotInfos[0], ShotInfos[1]);

			Keys.Ping = Keys.Keys;

			/*const ShotInfoList ShotInfos = *ShotInfosVariety.begin();//Darf keine Referenz sein, da wir unten die Daten brauchen, whrend wir das multiset ndern.

			//printf("Eigener Frame: Hash: %x\n", KIRam.GetHash());
			//printf("Move Key: %x\n", keys_packets[t%10].keys & 0x1f);
			bool DeleteFirst = ShotInfos[0].Time <= KIGame.CurrentTime;
			if(DeleteFirst || (int)ShotInfos.size() < (Game.RandomSeedDistribution.GetNumSetBits() == 1 ? FutureShots : 3))
			{
				KIGame.Move(keys_packets[t%10]);
				std::multiset<ShotInfoList> NewShotInfosVariety;
				int Index = 0;
				for(std::multiset<ShotInfoList>::iterator i = ShotInfosVariety.begin();
					i != ShotInfosVariety.end();
					i++)
				{
					if(DeleteFirst && !(*i).SameMovement(ShotInfos))
						continue;
					if(!DeleteFirst && (*i)[0].Time <= KIGame.CurrentTime-1)//Fr diesen Schuss ist es bereits zu spt
						continue;
					if(++Index == MaxSameShotInfos)
						break;
					ShotInfoList OldList = *i;/////////////////
					if(DeleteFirst)
						i->DeleteFirst();
					KIGame.CalculateNextShot(*i, 0, NewShotInfosVariety);
				}
				assert(!NewShotInfosVariety.empty());
				ShotInfosVariety = NewShotInfosVariety;
			}*/
		}
/*
		const ShotInfoList ShotInfos = *ShotInfosVariety.begin();//Darf keine Referenz sein, da wir unten die Daten brauchen, whrend wir das multiset ndern.
		//int Needed = abs(NeededRotation(ShotInfos[0].TargetDirection, ShotInfos[1].TargetDirection, ShotInfos[1].Dir));
		//if(Needed > MaxNeeded)
		//	printf("Neues Max bei %i um %i\n", MaxNeeded = Needed, ShotInfos[1].Time);
		pScore = &Score;
		gameram.Move(ShotInfos[0], ShotInfos[1]);
		pScore = 0;
		if(ShotInfos[0].Time == gameram.CurrentTime-1)
		{
			std::multiset<ShotInfoList> NewShotInfosVariety;
			int Index = 0;
			for(std::multiset<ShotInfoList>::iterator i = ShotInfosVariety.begin();
				i != ShotInfosVariety.end();
				i++)
			{
				if(!(*i).SameMovement(ShotInfos))
					continue;
				if(++Index == MaxSameShotInfos)
					break;
				//ShotInfoList OldList = *i;
				i->DeleteFirst();
				gameram.CalculateNextShot(*i, 0, NewShotInfosVariety);
			}
			assert(!NewShotInfosVariety.empty());
			ShotInfosVariety = NewShotInfosVariety;
		}
		RunningTime++;
		char Buffer[128];
		sprintf_s(Buffer, 128, "Sco: %i0 %i%%  Zeit: %i",
			int(Score), RunningTime*100/(60*60*5), RunningTime/60);
		SetWindowText(hwnd, Buffer);
		if(60*60*5 == RunningTime)
		{
			printf("Score nach 5 Minuten ohne Saucer mit Seed %02x: %i0\n", StartSeed, Score);
			//exit(0);
			gameram.StartNewGame();
			if(++StartSeed == 11)
			{
				StartSeed = 0;
				if(++FutureShots == -1)
					exit(0);
				printf("FutureShots: %i\n", FutureShots);
			}
			gameram.RandomSeed1 = StartSeed;
			RunningTime = 0;
			Score = 0;
			t = 0;
			ShotInfosVariety.clear();
			GameRamsHistory.clear();
		}



		keys_packets[t%10] = gameram.CalcDirInfos(&ShotInfos[t%10], 0, true, -1, 0, KeysPacket());
		if(Latency)
			gameram = OriginalRam;*/

		//keys_packets[t%10].fire(t%2 == 0);


#else
		keys_packets[t%10].clear();
		if(GetForegroundWindow() == hwnd)
		{
			if(GetAsyncKeyState('Q') < 0)
				keys_packets[t%10].left(true);
			if(GetAsyncKeyState('W') < 0)
				keys_packets[t%10].right(true);
			if(GetAsyncKeyState('I') < 0)
				keys_packets[t%10].thrust(true);
			if(GetAsyncKeyState('O') < 0)
				keys_packets[t%10].fire(true);
			if(GetAsyncKeyState(VK_SPACE) < 0)
				keys_packets[t%10].hyperspace(true);
		}
#endif

		LastFrame = Frame;

		/*int min_dist = 0x7fffffff;
		int min_dx = 0;
		int min_dy = 0;
		if (game.ship_present)
		{
			for (int i=0; i<game.nasteroids; ++i)
			{   // nchstgelegenen Asteroiden suchen
				int dx = game.asteroids[i].x - game.ship_x;
				while (dx < -512) dx += 1024; // dx normalisieren auf -512 ... 511
				while (dx > 511) dx -= 1024;
				int dy = game.asteroids[i].y - game.ship_y;
				while (dy < -384) dy += 768;  // dy normalisieren auf -384 ... 383
				while (dy > 383) dy -= 768;
				int dist = dx*dx+dy*dy;  // Quadrat des Abstands zu diesem Asteroiden
				switch (game.asteroids[i].sf)
				{	// Abstand um den ungefhren Radius des Asteroiden korrigieren
					case 0:  // groer Asteroid
						dist -= 40*40;
						break;
					case 15: // mittlerer Asteroid
						dist -= 20*20;
						break;
					case 14: // kleiner Asteroid
						dist -= 8*8;
						break;
				}
				if (dist < min_dist)
				{
					min_dist = dist;
					min_dx = dx;
					min_dy = dy;
				}
			}
			if (game.saucer_present)
			{
				int dx = game.saucer_x - game.ship_x;
				while (dx < -512) dx += 1024;
				while (dx > 511) dx -= 1024;
				int dy = game.saucer_y - game.ship_y;
				while (dy < -384) dy += 768;
				while (dy > 383) dy -= 768;
				int dist = dx*dx+dy*dy;
				switch (game.saucer_size)
				{	// Abstand um den ungefhren Radius des UFOs korrigieren
				case 15: // groes UFO
					dist -= 20*12;
					break;
				case 14: // kleines UFO
					dist -= 10*6;
					break;
				}
				if (dist < min_dist)
				{
					min_dist = dist;
					min_dx = dx;
					min_dy = dy;
				}
			}

			// Schiff in Richtung auf das nchstgelegene Objekt drehen
			// mathematisch wird hier das Kreuzprodukt aus den Vektoren 
			// ship_dx/y/0 und min_dx/y/0 berechnet
			if (game.ship_dx * min_dy - game.ship_dy * min_dx > 0)
				keys_packets[t%10].left(true);
			else
				keys_packets[t%10].right(true);

			if (min_dist < 30*30)  // Flucht, wenn Kollision unausweichlich
				keys_packets[t%10].hyperspace(true);

			if (min_dist > 300*300/100) // beschleunigen, wenn nichts in der Nhe
				keys_packets[t%10].thrust(true);

			if (t % 2 == 0)  // Feuerknopf drcken, so schnell es geht
				keys_packets[t%10].fire(true);
		}*/
		t++;
#endif
	}
}

void GameRam::Render(HDC hdc, int cx, int cy)
{
	SelectObject(hdc, GetStockObject(NULL_BRUSH));
	SelectObject(hdc, CreatePen(PS_SOLID, 0, RGB(255,255,0)));
	for(int Pass = 0; Pass < 2; Pass++)
	{
		for(int i = 0; i < NumObjects; i++)
		{
			if(Objects[i].Status==0)
				continue;
			if(Objects[i].Status < 0 && Pass != 0 || //Explosion im ersten Pass zeichnen
				Objects[i].Status >= 0 && Pass == 0)
				continue;
			int Size = 100;
			if(Objects[i].Status < 0)//Explosion
				Size = -Objects[i].Status;
			else if(i==PlayerIndex)
				Size = 12;
			else if(i==SaucerIndex && Objects[i].Status == 1)
				Size = -10;
			else if(i==SaucerIndex && Objects[i].Status == 2)
				Size = -20;
			else if(i<PlayerIndex && (Objects[i].Status&0x01))
				Size = 13;
			else if(i<PlayerIndex && (Objects[i].Status&0x02))
				Size = 20;
			else if(i<PlayerIndex && (Objects[i].Status&0x04))
				Size = 33;
			else if(i>SaucerIndex)
				Size = 4;
			else
				assert(0);
			for(int ax = 0; ax < Objects[i].cX; ax++)
			for(int ay = 0; ay < Objects[i].cY; ay++)
			{
				int x = int(Objects[i].PX[ax]) * cx / 0x2000;
				int y = int(Objects[i].PY[ay]) * cy / 0x1800;
				y = cy - y;
				if(Size > 0)
					Ellipse(hdc, x-Size*cx/1033, y-Size*cy/733, x+Size*cx/1033, y+Size*cy/733);
				else
					Rectangle(hdc, x+Size*cx/1033, y+Size*cy/733, x-Size*cx/1033, y-Size*cy/733);
				if(i==PlayerIndex)
				{
					MoveToEx(hdc, x, y, 0);
					LineTo(hdc, int(cos(float(ShipDirection)/255.f*6.283f)*20.0f)+x,
								int(sin(float(ShipDirection)/255.f*6.283f)*-20.0f)+y);
				}
				if(Objects[i].cX != 1 || Objects[i].cY != 1)
				{
					char Buffer[128];
					sprintf_s(Buffer, 128, "%ix%i", Objects[i].cX, Objects[i].cY);
					TextOut(hdc, x, y, Buffer, (int)strlen(Buffer));
				}
			}
		}
		//SelectObject(hdc, GetStockObject(WHITE_BRUSH));
		DeleteObject(SelectObject(hdc, GetStockObject(BLACK_PEN)));
	}

	for(std::multiset<ShotInfoList>::iterator j = ShotInfosVariety.begin();
		j != ShotInfosVariety.end();
		j++)
	{
		for(int i = 0; i < (int)j->size()-2; i++)
		{
			SelectObject(hdc, CreatePen(PS_SOLID, 0, RGB(i*255/int(j->size()-2), i*255/int(j->size()-2), i*255/int(j->size()-2))));
			assert((*j)[i].Time != MaxInt);
			int x = int((*j)[i].StartX) * cx / 0x2000;
			int y = int((*j)[i].StartY) * cy / 0x1800;
			y = cy - y;
			MoveToEx(hdc, x, y, 0);
			x = (*j)[i].TargetXEasy() * cx / 0x2000;
			y = (*j)[i].TargetYEasy() * cy / 0x1800;
			y = cy - y;
			LineTo(hdc, x, y);

			DeleteObject(SelectObject(hdc, GetStockObject(WHITE_PEN)));
		}
		break;
	}

	/*for(int x = 0; x < cx; x++)
	for(int y = x%2; y < cy; y+=2)
	{
		int xAst = x * 0x2000 / cx;
		int yAst = (cy-y) * 0x1800 / cy;
		for(int Object = SaucerIndex; Object >= 0; Object--)
		{
			if(Objects[Object].Status <= 0)
				continue;
			bool collision = AreObjectsNear(0x1F, xAst, yAst, 10,
											Object,
											Objects[Object].PositionX,
											Objects[Object].PositionY, Objects[Object].Status);
			if(collision)
				SetPixel(hdc, x, y, RGB(255,0,0));
		}
	}*/

	char Buffer[512];
	int len = sprintf_s(Buffer, 512,
		"Zeit: %i (%i:%02i) Score: %i.%03i Leben: %i Seeds: %i TimerBits: %i KI Shots: %i UFO: %i VRams: %i "
#ifdef USE_NETWORK_THREAD
		"NetDelay: %i",
#else
		"Blockierend",
#endif
		CurrentTime, (CurrentTime-GameStartedTime)/60/60, (CurrentTime-GameStartedTime)/60%60,
		Score/1000, Score%1000,
		NumShips,
		RandomSeedDistribution.GetNumSetBits(),
		TimerBitsKnown,
		ShotInfosVariety.begin()->size()-2,
		SaucerCountdown,
		VRamGames.size()
#ifdef USE_NETWORK_THREAD
		, NetDelay
#endif
		);
	TextOut(hdc, 0, 0, Buffer, len);
}


void Player::InterpretScreen(FramePacket &packet, GameStatus& game)
{
	unsigned short vector_ram[512];
	int dx, dy, sf, vx, vy, vz, vs;
	int v1x = 0;
	int v1y = 0;
	int shipdetect = 0;

	game.clear();

	/* Vektor-RAM in 16-Bit-Worte konvertieren. War in der ersten Version mal ein sportlicher
	Typecast: unsigned short *vector_ram = (unsigned short*)packet.vectorram;
	Das klappt aber nur auf Little-Endian-Systemen, daher besser portabel: */
	for(int i=0; i<512; ++i)
//		vector_ram[i] = (unsigned char)packet.vectorram[2*i] | (unsigned char)packet.vectorram[2*i+1] << 8;
		vector_ram[i] = packet.VRam[i];

	if (vector_ram[0] != 0xe001 && vector_ram[0] != 0xe201)
		return; // sollte nicht vorkommen; erster Befehl ist immer ein JMPL

	int pc = 1;
	while (pc < 512)
	{
		int op = vector_ram[pc] >> 12;
		switch (op)
		{
		case 0xa: // LABS
			vy = vector_ram[pc] & 0x3ff;
			vx = vector_ram[pc+1] & 0x3ff;
			vs = vector_ram[pc+1] >> 12;
			break;
		case 0xb: // HALT
			return;
		case 0xc: // JSRL
			switch (vector_ram[pc] & 0xfff)
			{
			case 0x8f3:
				game.asteroids[game.nasteroids++].set(vx, vy, 1, vs);
				break;
			case 0x8ff:
				game.asteroids[game.nasteroids++].set(vx, vy, 2, vs);
				break;
			case 0x90d:
				game.asteroids[game.nasteroids++].set(vx, vy, 3, vs);
				break;
			case 0x91a:
				game.asteroids[game.nasteroids++].set(vx, vy, 4, vs);
				break;
			case 0x929:
				game.saucer_present = true;
				game.saucer_x = vx;
				game.saucer_y = vy;
				game.saucer_size = vs;
				break;
			}  
			break;
		case 0xd: // RTSL
			return;
		case 0xe: // JMPL
			/*
			pc = vector_ram[pc] & 0xfff;
			break;
			*/
			return;
		case 0xf: // SVEC
			/*
			dy = vector_ram[pc] & 0x300;
			if ((vector_ram[pc] & 0x400) != 0)
				dy = -dy;
			dx = (vector_ram[pc] & 3) << 8;
			if ((vector_ram[pc] & 4) != 0)
				dx = -dx;
			sf = (((vector_ram[pc] & 8) >> 2) | ((vector_ram[pc] & 0x800) >> 11)) + 2;
			vz = (vector_ram[pc] & 0xf0) >> 4;
			*/
			break;
		default:
			dy = vector_ram[pc] & 0x3ff;
			if(vector_ram[pc] & 0x400)
				dy = -dy;
			dx = vector_ram[pc+1] & 0x3ff;
			if(vector_ram[pc+1] & 0x400)
				dx = -dx;
			sf = op;
			vz = vector_ram[pc+1] >> 12;
			if (dx == 0 && dy == 0 && vz == 15)
				game.shots[game.nshots++].set(vx, vy);
			if (op == 6 && vz == 12 && dx != 0 && dy != 0)
			{
				switch (shipdetect)
				{
				case 0:
					v1x = dx;
					v1y = dy;
					++shipdetect;
					break;
				case 1:
					game.ship_present = true;
					game.ship_x = vx;
					game.ship_y = vy;
					game.ship_dx = v1x - dx;
					game.ship_dy = v1y - dy;
					++shipdetect;
					break;
				}
			}
			else if (shipdetect == 1)
				shipdetect = 0;

			break;
		}
		if (op <= 0xa)
			++pc;
		if (op != 0xe) // JMPL
			++pc;
	}   

}


void Asteroid::set(int x, int y, int type, int sf)
{
	this->x = x;
	this->y = y;
	this->type = type;
	this->sf = sf;
}

void Shot::set(int x, int y)
{
	this->x = x;
	this->y = y;
}

void GameStatus::clear(void)
{
	ship_present = false;
	saucer_present = false;
	nasteroids = 0;
	nshots = 0;
}
/*
KeysPacket::KeysPacket(void)
{
	signature[0] = 'c';
	signature[1] = 't';
	signature[2] = 'm';
	signature[3] = 'a';
	signature[4] = 'm';
	signature[5] = 'e';
	keys = 0;
	Ping = 0;
}

void KeysPacket::clear(void)
{
	keys = '@';
}

void KeysPacket::hyperspace(bool b)
{
	if (b)
		keys |= KEY_HYPERSPACE;
	else
		keys &= ~KEY_HYPERSPACE;
}

void KeysPacket::fire(bool b)
{
	if (b)
		keys |= KEY_FIRE;
	else
		keys &= ~KEY_FIRE;
}

void KeysPacket::thrust(bool b)
{
	if (b)
		keys |= KEY_THRUST;
	else
		keys &= ~KEY_THRUST;
}

void KeysPacket::left(bool b)
{
	if (b)
	{
		keys |= KEY_LEFT;
//		right(false);
	}
	else
		keys &= ~KEY_LEFT;
}

void KeysPacket::right(bool b)
{
	if (b)
	{
		keys |= KEY_RIGHT;
//		left(false);
	}
	else
		keys &= ~KEY_RIGHT;
}
*/
/*
void Player::ReceivePacket(FramePacket &packet)
{
	sockaddr_in sender;
	int sender_size = sizeof sender;
	fd_set readfds, writefds, exceptfds;

	do
	{
		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		FD_ZERO(&exceptfds);
		FD_SET(sd, &readfds);
		FD_SET(sd, &exceptfds);
		select(int(sd)+1, &readfds, &writefds, &exceptfds, NULL);
		int bytes_received = recv(sd, (char *)&packet, sizeof packet, 0);
		if (bytes_received != sizeof packet)
		{
			int err = WSAGetLastError();
			fprintf(stderr, "Fehler %d bei recvfrom().\n", err);
			exit(0);
		}
		break;
		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		FD_ZERO(&exceptfds);
		FD_SET(sd, &readfds);
		timeval zero;
		zero.tv_sec = zero.tv_usec = 0;
		select(int(sd)+1, &readfds, &writefds, &exceptfds, &zero);
		if(FD_ISSET(sd, &readfds))
			printf("Paket weggeschmissen\n");
	} while(FD_ISSET(sd, &readfds));
}

void Player::SendPacket(KeysPacket &packet)
{
	sockaddr_in server;
	memset(&server, 0, sizeof server);
	server.sin_family = AF_INET;
	server.sin_port = htons(1979);
	server.sin_addr.s_addr = server_ip;
	if (sizeof packet != sendto(sd, (char *)&packet, sizeof packet, 0, (sockaddr*)&server, sizeof server))
	{
		int err = WSAGetLastError();
		if (err != WSAEWOULDBLOCK)
		{
			fprintf(stderr, "Fehler %d bei sendto().\n", err);
			exit(1);
		}
	}
}*/

#ifdef INCLUDE_WINDOW
LRESULT CALLBACK WndProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static int xView=0, yView=0;
	
	switch (message)
	{
	case WM_CREATE:
	{
		::hwnd = hwnd;
		HDC hdcScreen = GetDC(hwnd);
		hBitmap = CreateCompatibleBitmap(hdcScreen, xChars*xFont, yChars*yFont);
		hdcMem  = CreateCompatibleDC(hdcScreen);
		SelectObject(hdcMem, hBitmap);
		ReleaseDC(hwnd, hdcScreen);

		SelectObject(hdcMem, CreateFont (0, 0, 0, 0, 0, 0, 0, 0, DEFAULT_CHARSET, 0, 0, 0, FIXED_PITCH, NULL));
		Rectangle(hdcMem, -1, -1, xChars*xFont+1, yChars*yFont+1);
		for(int i=0; i<0x100; i++)
		{
			char Buffer[3];
			sprintf_s(Buffer, 3, "%2.2x", i);
			TextOut(hdcMem, i*xFont*2, 0, Buffer, 2);
		}
/*		TEXTMETRIC tm;
		GetTextMetrics(hdcMem, &tm);
		xFont = tm.tmAveCharWidth;
		yFont = tm.tmHeight;*/
		return 0;
	}

	case WM_SIZE:
		cxClient = LOWORD(lParam);
		cyClient = HIWORD(lParam);
		while(xView+cxClient/xFont>xChars)
			xView--;
		return 0 ;

	case WM_KEYDOWN:
#ifdef USE_NETWORK_THREAD
		if(wParam == 'Y')
			NetDelay++;
		else if(wParam == 'X')
			NetDelay = Max(NetDelay-1, 0);
#endif
		if(wParam==VK_LEFT)
		{
			xView-=2;
			CurrentGameRam--;
			if(CurrentGameRam < 0)
				CurrentGameRam = (int)GameRamsHistory.size()-1;
		}
		else if(wParam==VK_RIGHT)
		{
			xView+=2;
			CurrentGameRam++;
			if(CurrentGameRam >= (int)GameRamsHistory.size())
				CurrentGameRam = (int)GameRamsHistory.size()-1;
		}
		else if(wParam==VK_UP)
		{
			yView--;
			CurrentGameRam = -1;
		}
		else if(wParam==VK_DOWN)
		{
			yView++;
//			ShotInfo S;
//			while(CurrentGameRam != -1)
//				GameRamsHistory[CurrentGameRam].CalcDirInfos(&S, 0, true, -1, 0, KeysPacket());
		}
		else
			return 0;
		if(xView<0) xView=0;
		if(yView<0) yView=0;
		while(xView+cxClient/xFont>xChars) xView--;
		InvalidateRect(hwnd, 0, false);
		return 0;

	/*case WM_LBUTTONDOWN:
		updating = !updating;
		if(!updating)
		{
			gameram.StartNewGame();
			RunningTime = 0;
		}
		return 0;*/

	case WM_PAINT:
	{
		PAINTSTRUCT	ps;
		HDC hdcScreen = BeginPaint(hwnd, &ps);

		HBITMAP hBitmap = CreateCompatibleBitmap(hdcScreen, cxClient, cyClient);
		HDC hdc = CreateCompatibleDC(hdcScreen);
		SelectObject(hdc, hBitmap);

		RECT rc;
		SetRect(&rc, 0, 0, cxClient, cyClient);
		FillRect(hdc, &rc, (HBRUSH)GetStockObject(WHITE_BRUSH));

		//BitBlt(hdc, 0, 0, cxClient, cyClient,
		//       hdcMem, xView*xFont, yView*yFont, SRCCOPY);
		if(CurrentGameRam == -1)
		{
			VRamGames[0].Render(hdc, cxClient, cyClient);
		}
		else
		{
			//ShotInfo S;
			//GameRamsHistory[CurrentGameRam].CalcDirInfos(&S, 0, true, -1, 0, KeysPacket());
			GameRamsHistory[CurrentGameRam].Render(hdc, cxClient, cyClient);
		}

		BitBlt(hdcScreen, 0, 0, cxClient, cyClient, hdc, 0, 0, SRCCOPY);
		DeleteDC(hdc);
		DeleteObject(hBitmap);
		EndPaint(hwnd, &ps);
		return 0;

	}
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc (hwnd, message, wParam, lParam) ;
}

#endif